@@ -1,7 +1,7 @@
 /*
  * This file is part of GPaste.
  *
- * Copyright (c) 2010-2016, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
+ * Copyright (c) 2010-2017, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
  */
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
@@ -12,6 +12,7 @@ const St = imports.gi.St;
 
 const GPastePageItem = new Lang.Class({
     Name: 'GPastePageItem',
+    Implements: [Signals.WithSignals],
 
     _init: function(page) {
         this.actor = new St.Button({
@@ -44,4 +45,3 @@ const GPastePageItem = new Lang.Class({
         this.actor.destroy();
     }
 });
-Signals.addSignalMethods(GPastePageItem.prototype);
