@@ -1,12 +1,13 @@
 /*
  * This file is part of GPaste.
  *
- * Copyright (c) 2010-2016, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
+ * Copyright (c) 2010-2017, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
  */
 
 #include <gpaste-history.h>
 #include <gpaste-image-item.h>
 #include <gpaste-gsettings-keys.h>
+#include <gpaste-storage-backend.h>
 #include <gpaste-update-enums.h>
 #include <gpaste-uris-item.h>
 #include <gpaste-util.h>
@@ -25,17 +26,18 @@ enum
 
 typedef struct
 {
-    GPasteSettings *settings;
-    GList          *history;
-    guint64         size;
+    GPasteStorageBackend *backend;
+    GPasteSettings       *settings;
+    GList                *history;
+    guint64               size;
 
-    gchar          *name;
+    gchar                *name;
 
     /* Note: we never track the first (active) item here */
-    guint64         biggest_index;
-    guint64         biggest_size;
+    guint64               biggest_index;
+    guint64               biggest_size;
 
-    guint64         c_signals[C_LAST_SIGNAL];
+    guint64               c_signals[C_LAST_SIGNAL];
 } GPasteHistoryPrivate;
 
 G_PASTE_DEFINE_TYPE_WITH_PRIVATE (History, history, G_TYPE_OBJECT)
@@ -355,6 +357,39 @@ g_paste_history_remove (GPasteHistory *self,
     g_paste_history_update (self, G_PASTE_UPDATE_ACTION_REMOVE, G_PASTE_UPDATE_TARGET_POSITION, pos);
 }
 
+/**
+ * g_paste_history_refresh_item_size:
+ * @self: a #GPasteHistory instance
+ * @item: the #GPasteItem to refresh
+ * @old_size: the former size of the item
+ *
+ * Refresh the cached size of the #GPasteItem
+ */
+G_PASTE_VISIBLE void
+g_paste_history_refresh_item_size (GPasteHistory    *self,
+                                   const GPasteItem *item,
+                                   guint64           old_size)
+{
+    g_return_if_fail (_G_PASTE_IS_HISTORY (self));
+    g_return_if_fail (_G_PASTE_IS_ITEM (item));
+
+    GPasteHistoryPrivate *priv = g_paste_history_get_instance_private (self);
+
+    if (!g_list_find (priv->history, item))
+        return;
+
+    guint64 size = g_paste_item_get_size (item);
+
+    g_return_if_fail (old_size <= size);
+
+    priv->size += (size - old_size);
+
+    if (size > priv->biggest_size)
+        g_paste_history_private_elect_new_biggest (priv);
+
+    g_paste_history_private_check_memory_usage (priv);
+}
+
 static GPasteItem *
 g_paste_history_private_get (const GPasteHistoryPrivate *priv,
                              guint64                     pos)
@@ -532,6 +567,30 @@ g_paste_history_replace (GPasteHistory *self,
         g_paste_history_selected (self, new);
 }
 
+static GPasteItem *
+_g_paste_history_private_get_password (const GPasteHistoryPrivate *priv,
+                                       const gchar                *name,
+                                       guint64                    *index)
+{
+    guint64 idx = 0;
+
+    for (GList *h = priv->history; h; h = g_list_next (h), ++idx)
+    {
+        GPasteItem *i = h->data;
+        if (_G_PASTE_IS_PASSWORD_ITEM (i) &&
+            g_paste_str_equal (g_paste_password_item_get_name ((GPastePasswordItem *) i), name))
+        {
+            if (index)
+                *index = idx;
+            return i;
+        }
+    }
+
+    if (index)
+        *index = -1;
+    return NULL;
+}
+
 /**
  * g_paste_history_set_password:
  * @self: a #GPasteHistory instance
@@ -560,36 +619,13 @@ g_paste_history_set_password (GPasteHistory *self,
     GPasteItem *item = todel->data;
 
     g_return_if_fail (_G_PASTE_IS_TEXT_ITEM (item) && g_paste_str_equal (g_paste_item_get_kind (item), "Text"));
+    g_return_if_fail (!_g_paste_history_private_get_password (priv, name, NULL));
 
     GPasteItem *password = g_paste_password_item_new (name, g_paste_item_get_real_value (item));
 
     _g_paste_history_replace (self, index, password, todel);
 }
 
-static GPasteItem *
-_g_paste_history_private_get_password (const GPasteHistoryPrivate *priv,
-                                       const gchar                *name,
-                                       guint64                    *index)
-{
-    guint64 idx = 0;
-
-    for (GList *h = priv->history; h; h = g_list_next (h), ++idx)
-    {
-        GPasteItem *i = h->data;
-        if (_G_PASTE_IS_PASSWORD_ITEM (i) &&
-            g_paste_str_equal (g_paste_password_item_get_name ((GPastePasswordItem *) i), name))
-        {
-            if (index)
-                *index = idx;
-            return i;
-        }
-    }
-
-    if (index)
-        *index = -1;
-    return NULL;
-}
-
 /**
  * g_paste_history_get_password:
  * @self: a #GPasteHistory instance
@@ -681,77 +717,6 @@ g_paste_history_empty (GPasteHistory *self)
     g_paste_history_update (self, G_PASTE_UPDATE_ACTION_REMOVE, G_PASTE_UPDATE_TARGET_ALL, 0);
 }
 
-static gchar *
-g_paste_history_encode (const gchar *text)
-{
-    g_autofree gchar *_encoded_text = g_paste_util_replace (text, "&", "&amp;");
-    return g_paste_util_replace (_encoded_text, ">", "&gt;");
-}
-
-static gchar *
-g_paste_history_decode (const gchar *text)
-{
-    g_autofree gchar *_decoded_text = g_paste_util_replace (text, "&gt;", ">");
-    return g_paste_util_replace (_decoded_text, "&amp;", "&");
-}
-
-static gchar *
-g_paste_history_get_history_dir_path (void)
-{
-    return g_build_filename (g_get_user_data_dir (), "gpaste", NULL);
-}
-
-static GFile *
-g_paste_history_get_history_dir (void)
-{
-    g_autofree gchar *history_dir_path = g_paste_history_get_history_dir_path ();
-    return g_file_new_for_path (history_dir_path);
-}
-
-static gchar *
-g_paste_history_get_history_file_path (const gchar *name)
-{
-    g_return_val_if_fail (name, NULL);
-
-    g_autofree gchar *history_dir_path = g_paste_history_get_history_dir_path ();
-    g_autofree gchar *history_file_name = g_strconcat (name, ".xml", NULL);
-
-    return g_build_filename (history_dir_path, history_file_name, NULL);
-}
-
-static GFile *
-g_paste_history_get_history_file (const gchar *name)
-{
-    g_autofree gchar *history_file_path = g_paste_history_get_history_file_path (name);
-    return g_file_new_for_path (history_file_path);
-}
-
-static gboolean
-ensure_history_dir_exists (gboolean save_history)
-{
-    g_autoptr (GFile) history_dir = g_paste_history_get_history_dir ();
-
-    if (!g_file_query_exists (history_dir,
-                              NULL)) /* cancellable */
-    {
-        if (!save_history)
-            return TRUE;
-
-        g_autoptr (GError) error = NULL;
-
-        g_file_make_directory_with_parents (history_dir,
-                                            NULL, /* cancellable */
-                                            &error);
-        if (error)
-        {
-            g_critical ("%s: %s", _("Could not create history dir"), error->message);
-            return FALSE;
-        }
-    }
-
-    return TRUE;
-}
-
 /**
  * g_paste_history_save:
  * @self: a #GPasteHistory instance
@@ -767,282 +732,9 @@ g_paste_history_save (GPasteHistory *self,
 
     const GPasteHistoryPrivate *priv = _g_paste_history_get_instance_private (self);
 
-    GPasteSettings *settings = priv->settings;
-    gboolean save_history = g_paste_settings_get_save_history (settings);
-    g_autofree gchar *history_file_path = NULL;
-    g_autoptr (GFile) history_file = NULL;
-
-    if (!ensure_history_dir_exists (save_history))
-        return;
-
-    history_file_path = g_paste_history_get_history_file_path ((name) ? name : priv->name);
-    history_file = g_file_new_for_path (history_file_path);
-
-    if (!save_history)
-    {
-        g_file_delete (history_file,
-                       NULL, /* cancellable*/
-                       NULL); /* error */
-    }
-    else
-    {
-        g_autoptr (GOutputStream) stream = G_OUTPUT_STREAM (g_file_replace (history_file,
-                                                            NULL,
-                                                            FALSE,
-                                                            G_FILE_CREATE_REPLACE_DESTINATION,
-                                                            NULL, /* cancellable */
-                                                            NULL)); /* error */
-
-        if (!g_output_stream_write_all (stream, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", 39, NULL, NULL /* cancellable */, NULL /* error */) ||
-            !g_output_stream_write_all (stream, "<history version=\"1.0\">\n", 24, NULL, NULL /* cancellable */, NULL /* error */))
-                return;
-
-        for (GList *history = priv->history; history; history = g_list_next (history))
-        {
-            GPasteItem *item = history->data;
-            const gchar *kind = g_paste_item_get_kind (item);
-
-            if (g_paste_str_equal (kind, "Password"))
-                continue;
-
-            g_autofree gchar *text = g_paste_history_encode (g_paste_item_get_value (item));
-
-            if (!g_output_stream_write_all (stream, "  <item kind=\"", 14, NULL, NULL /* cancellable */, NULL /* error */) ||
-                !g_output_stream_write_all (stream, kind, strlen (kind), NULL, NULL /* cancellable */, NULL /* error */) ||
-                (_G_PASTE_IS_IMAGE_ITEM (item) &&
-                    (!g_output_stream_write_all (stream, "\" date=\"", 8, NULL, NULL /* cancellable */, NULL /* error */) ||
-                     !g_output_stream_write_all (stream, g_date_time_format ((GDateTime *) g_paste_image_item_get_date (G_PASTE_IMAGE_ITEM (item)), "%s"), 10, NULL, NULL /* cancellable */, NULL /* error */))) ||
-                !g_output_stream_write_all (stream, "\"><![CDATA[", 11, NULL, NULL /* cancellable */, NULL /* error */) ||
-                !g_output_stream_write_all (stream, text, strlen (text), NULL, NULL /* cancellable */, NULL /* error */) ||
-                !g_output_stream_write_all (stream, "]]></item>\n", 11, NULL, NULL /* cancellable */, NULL /* error */))
-            {
-                g_warning ("Failed to write an item to history");
-                continue;
-            }
-        }
-
-        if (!g_output_stream_write_all (stream, "</history>\n", 11, NULL, NULL /* cancellable */, NULL /* error */) ||
-            !g_output_stream_close (stream, NULL /* cancellable */, NULL /* error */))
-                return;
-    }
+    g_paste_storage_backend_write_history (priv->backend, (name) ? name : priv->name, priv->history);
 }
 
-/********************/
-/* Begin XML Parser */
-/********************/
-
-typedef enum
-{
-    BEGIN,
-    IN_HISTORY,
-    IN_ITEM,
-    HAS_TEXT,
-    END
-} State;
-
-typedef enum
-{
-    TEXT,
-    IMAGE,
-    URIS,
-    PASSWORD
-} Type;
-
-typedef struct
-{
-    GPasteHistoryPrivate *priv;
-    State                 state;
-    Type                  type;
-    guint64               current_size;
-    guint64               max_size;
-    gboolean              images_support;
-    gchar                *date;
-    gchar                *name;
-    gchar                *text;
-} Data;
-
-#define ASSERT_STATE(x)                                                                           \
-    if (data->state != x)                                                                         \
-    {                                                                                             \
-        g_warning ("Expected state %" G_GINT32_FORMAT ", but got %" G_GINT32_FORMAT, x, data->state); \
-        return;                                                                                   \
-    }
-#define SWITCH_STATE(x, y) \
-    ASSERT_STATE (x);      \
-    data->state = y
-
-static void
-start_tag (GMarkupParseContext *context G_GNUC_UNUSED,
-           const gchar         *element_name,
-           const gchar        **attribute_names,
-           const gchar        **attribute_values,
-           gpointer             user_data,
-           GError             **error G_GNUC_UNUSED)
-{
-    Data *data = user_data;
-
-    if (g_paste_str_equal (element_name, "history"))
-    {
-        SWITCH_STATE (BEGIN, IN_HISTORY);
-    }
-    else if (g_paste_str_equal (element_name, "item"))
-    {
-        SWITCH_STATE (IN_HISTORY, IN_ITEM);
-        g_clear_pointer (&data->date, g_free);
-        g_clear_pointer (&data->text, g_free);
-        for (const gchar **a = attribute_names, **v = attribute_values; *a && *v; ++a, ++v)
-        {
-            if (g_paste_str_equal (*a, "kind"))
-            {
-                if (g_paste_str_equal (*v, "Text"))
-                    data->type = TEXT;
-                else if (g_paste_str_equal (*v, "Image"))
-                    data->type = IMAGE;
-                else if (g_paste_str_equal (*v, "Uris"))
-                    data->type = URIS;
-                else if (g_paste_str_equal (*v, "Password"))
-                    data->type = PASSWORD;
-                else
-                    g_critical ("Unknown item kind: %s", *v);
-            }
-            else if (g_paste_str_equal (*a, "date"))
-            {
-                if (data->type != IMAGE)
-                {
-                    g_warning ("Expected type %" G_GINT32_FORMAT ", but got %" G_GINT32_FORMAT, IMAGE, data->type);
-                    return;
-                }
-                data->date = g_strdup (*v);
-            }
-            else if (g_paste_str_equal (*a, "name"))
-            {
-                if (data->type != PASSWORD)
-                {
-                    g_warning ("Expected type %" G_GINT32_FORMAT ", but got %" G_GINT32_FORMAT, PASSWORD, data->type);
-                    return;
-                }
-                data->name = g_strdup (*v);
-            }
-            else
-                g_warning ("Unknown item attribute: %s", *a);
-        }
-    }
-    else
-        g_warning ("Unknown element: %s", element_name);
-}
-
-static void
-end_tag (GMarkupParseContext *context G_GNUC_UNUSED,
-         const gchar         *element_name,
-         gpointer             user_data,
-         GError             **error G_GNUC_UNUSED)
-{
-    Data *data = user_data;
-
-    if (g_paste_str_equal (element_name, "history"))
-    {
-        SWITCH_STATE (IN_HISTORY, END);
-    }
-    else if (g_paste_str_equal (element_name, "item"))
-    {
-        SWITCH_STATE (HAS_TEXT, IN_HISTORY);
-    }
-    else
-        g_warning ("Unknown element: %s", element_name);
-}
-
-static void
-on_text (GMarkupParseContext *context G_GNUC_UNUSED,
-         const gchar         *text,
-         guint64              text_len,
-         gpointer             user_data,
-         GError             **error G_GNUC_UNUSED)
-{
-    Data *data = user_data;
-
-    g_autofree gchar *txt = g_strndup (text, text_len);
-    switch (data->state)
-    {
-    case IN_HISTORY:
-    case HAS_TEXT:
-        if (*g_strstrip (txt))
-        {
-            g_warning ("Unexpected text: %s", txt);
-            return;
-        }
-        break;
-    case IN_ITEM:
-    {
-        g_autofree gchar *value = g_paste_history_decode (txt);
-        if (*g_strstrip (txt))
-        {
-            if (data->current_size < data->max_size)
-            {
-                GPasteItem *item = NULL;
-
-                switch (data->type)
-                {
-                case TEXT:
-                    item = g_paste_text_item_new (value);
-                    break;
-                case URIS:
-                    item = g_paste_uris_item_new (value);
-                    break;
-                case PASSWORD:
-                    item = g_paste_password_item_new (data->name, value);
-                    break;
-                case IMAGE:
-                    if (data->images_support && data->date)
-                    {
-                        g_autoptr (GDateTime) date_time = g_date_time_new_from_unix_local (g_ascii_strtoll (data->date,
-                                                                                                            NULL, /* end */
-                                                                                                            0)); /* base */
-                        item = g_paste_image_item_new_from_file (value, date_time);
-                    }
-                    else
-                    {
-                        g_autoptr (GFile) img_file = g_file_new_for_path (value);
-
-                        if (g_file_query_exists (img_file,
-                                                 NULL)) /* cancellable */
-                        {
-                            g_file_delete (img_file,
-                                           NULL, /* cancellable */
-                                           NULL); /* error */
-                        }
-                    }
-                    break;
-                }
-
-                if (item)
-                {
-                    GPasteHistoryPrivate *priv = data->priv;
-                    priv->size += g_paste_item_get_size (item);
-                    priv->history = g_list_append (priv->history, item);
-                    ++data->current_size;;
-                }
-            }
-
-            SWITCH_STATE (IN_ITEM, HAS_TEXT);
-        }
-        break;
-    }
-    default:
-        g_warning ("Unexpected state: %" G_GINT32_FORMAT, data->state);
-    }
-}
-
-static void on_error (GMarkupParseContext *context   G_GNUC_UNUSED,
-                      GError              *error,
-                      gpointer             user_data G_GNUC_UNUSED)
-{
-    g_critical ("error: %s", error->message);
-}
-
-/******************/
-/* End XML Parser */
-/******************/
-
 /**
  * g_paste_history_load:
  * @self: a #GPasteHistory instance
@@ -1058,7 +750,6 @@ g_paste_history_load (GPasteHistory *self,
     g_return_if_fail (!name || g_utf8_validate (name, -1, NULL));
 
     GPasteHistoryPrivate *priv = g_paste_history_get_instance_private (self);
-    GPasteSettings *settings = priv->settings;
 
     if (priv->name && g_paste_str_equal(name, priv->name))
         return;
@@ -1066,55 +757,12 @@ g_paste_history_load (GPasteHistory *self,
     g_list_free_full (priv->history,
                       g_object_unref);
     priv->history = NULL;
+    priv->size = 0;
 
     g_free (priv->name);
     priv->name = g_strdup ((name) ? name : g_paste_settings_get_history_name (priv->settings));
 
-    g_autofree gchar *history_file_path = g_paste_history_get_history_file_path (priv->name);
-    g_autoptr (GFile) history_file = g_file_new_for_path (history_file_path);
-
-    if (g_file_query_exists (history_file,
-                             NULL)) /* cancellable */
-    {
-        GMarkupParser parser = {
-            start_tag,
-            end_tag,
-            on_text,
-            NULL,
-            on_error
-        };
-        Data data = {
-            priv,
-            BEGIN,
-            TEXT,
-            0,
-            g_paste_settings_get_max_history_size (settings),
-            g_paste_settings_get_images_support (settings),
-            NULL,
-            NULL,
-            NULL
-        };
-        GMarkupParseContext *ctx = g_markup_parse_context_new (&parser,
-                                                               G_MARKUP_TREAT_CDATA_AS_TEXT,
-                                                               &data,
-                                                               NULL);
-        gchar *text;
-        guint64 text_length;
-
-        g_file_get_contents (history_file_path, &text, &text_length, NULL);
-        g_markup_parse_context_parse (ctx, text, text_length, NULL);
-        g_markup_parse_context_end_parse (ctx, NULL);
-
-        if (data.state != END)
-            g_warning ("Unexpected state adter parsing history: %" G_GINT32_FORMAT, data.state);
-        g_markup_parse_context_unref (ctx);
-    }
-    else
-    {
-        /* Create the empty file to be listed as an available history */
-        if (ensure_history_dir_exists (g_paste_settings_get_save_history (settings)))
-            g_object_unref (g_file_create (history_file, G_FILE_CREATE_NONE, NULL, NULL));
-    }
+    g_paste_storage_backend_read_history (priv->backend, priv->name, &priv->history, &priv->size);
 
     if (priv->history)
     {
@@ -1160,7 +808,7 @@ g_paste_history_delete (GPasteHistory *self,
 
     const GPasteHistoryPrivate *priv = _g_paste_history_get_instance_private (self);
 
-    g_autoptr (GFile) history_file = g_paste_history_get_history_file ((name) ? name : priv->name);
+    g_autoptr (GFile) history_file = g_paste_util_get_history_file ((name) ? name : priv->name, "xml");
 
     if (g_paste_str_equal (name, priv->name))
         g_paste_history_empty (self);
@@ -1210,6 +858,8 @@ g_paste_history_dispose (GObject *object)
     const GPasteHistoryPrivate *priv = _g_paste_history_get_instance_private (self);
     GPasteSettings *settings = priv->settings;
 
+    g_clear_object (&priv->backend);
+
     if (settings)
     {
         g_signal_handler_disconnect (settings, priv->c_signals[C_CHANGED]);
@@ -1453,6 +1103,7 @@ g_paste_history_new (GPasteSettings *settings)
     GPasteHistory *self = g_object_new (G_PASTE_TYPE_HISTORY, NULL);
     GPasteHistoryPrivate *priv = g_paste_history_get_instance_private (self);
 
+    priv->backend = g_paste_storage_backend_new (G_PASTE_STORAGE_DEFAULT, settings);
     priv->settings = g_object_ref (settings);
     priv->c_signals[C_CHANGED] = g_signal_connect (settings,
                                                    "changed",
@@ -1479,7 +1130,7 @@ g_paste_history_list (GError **error)
     g_autoptr (GArray) history_names = g_array_new (TRUE, /* zero-terminated */
                                                     TRUE, /* clear */
                                                     sizeof (gchar *));
-    g_autoptr (GFile) history_dir = g_paste_history_get_history_dir ();
+    g_autoptr (GFile) history_dir = g_paste_util_get_history_dir ();
     g_autoptr (GFileEnumerator) histories = g_file_enumerate_children (history_dir,
                                                                        G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
                                                                        G_FILE_QUERY_INFO_NONE,
