@@ -1,7 +1,7 @@
 /*
  * This file is part of GPaste.
  *
- * Copyright (c) 2010-2016, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
+ * Copyright (c) 2010-2017, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
  */
 
 #include <gpaste-ui-search.h>
@@ -25,6 +25,7 @@ g_paste_ui_search_init (GPasteUiSearch *self)
 
     gtk_widget_set_tooltip_text (widget, _("Search"));
     gtk_widget_set_valign (widget, GTK_ALIGN_CENTER);
+    gtk_container_add (GTK_CONTAINER (self), gtk_image_new_from_icon_name ("edit-find-symbolic", GTK_ICON_SIZE_BUTTON));
 }
 
 /**
@@ -38,7 +39,5 @@ g_paste_ui_search_init (GPasteUiSearch *self)
 G_PASTE_VISIBLE GtkWidget *
 g_paste_ui_search_new (void)
 {
-    return gtk_widget_new (G_PASTE_TYPE_UI_SEARCH,
-                           "image", gtk_image_new_from_icon_name ("edit-find-symbolic", GTK_ICON_SIZE_BUTTON),
-                           NULL);
+    return gtk_widget_new (G_PASTE_TYPE_UI_SEARCH, NULL);
 }
